function [params, LLhist, errhist] = ApproxEMforLDS( y_, u_, iparams, priors, numiters, updateTransMat, updateCovMat, updateInit, klim, sampleperiod, method )

if nargin >= 11
    LLhist = zeros(numiters/sampleperiod, 1);
    errhist = zeros(numiters/sampleperiod, 1);
end

[A,B,C,D,Q,R,insize,outsize,hidsize,pi_1,V_1] = pextract( iparams );

T = length(y_);

y_y_0 = y_*y_';
y_u_0 = y_*u_';
u_u_0 = u_*u_';

u_u_0_2 = u_u_0 - u_(:,end)*u_(:,end)';
inv_u_u_0 = inv(u_u_0);
inv_u_u_0_2 = inv(u_u_0_2);


if ~isempty( priors )
    error( 'priors not implemented' );
    %[sigA, nQ, invVQ, nR, invVR] = rextract( priors );
end

ape = ApproxEStep( y_, u_, klim, 2*klim+1, 25, insize, outsize );

%start of loop:
for iter = 1:numiters
    
    disp( ['Iteration ' num2str(iter) ':'] );
    
    [ape, expt, err, LL] = Step( ape, pstruct( A, B, C, D, Q, R, pi_1, V_1 ) );

%% Compute approximate E

    disp( ['Approximate LL: ' num2str(LL)]);
    disp( ['Approximate average prediction error: ', num2str(sqrt(err/T)) ]);        
    

%% M-Step
    [Ex_x_0, Ex_x_1, Ey_x_0, Eu_x_0, Ex_u_1, Ex_, Exx_] = eextract( expt );
    
    newC = (Ey_x_0 - y_u_0*inv_u_u_0*Eu_x_0)*inv( Ex_x_0 - Eu_x_0'*inv_u_u_0*Eu_x_0 );
    newD = (y_u_0 - newC*Eu_x_0')*inv_u_u_0;
 
%     % static D version:
%     newC = Ey_x_0*inv(Ex_x_0);
%     newD = D;


%    if iter > 350
        Eu_x_0x = Eu_x_0 - u_(:,T)*Ex_.end';
        newA = (Ex_x_1 - Ex_u_1*inv_u_u_0_2*Eu_x_0x)*inv((Ex_x_0 - Exx_.end) - Eu_x_0x'*inv_u_u_0_2*Eu_x_0x);
        newB = (Ex_u_1 - newA*Eu_x_0x')*inv_u_u_0_2;
%    else
%        % static B version:
%        newA = (Ex_x_1 - B*Eu_x_1)*inv(Ex_x_0 - Exx_(:,T));
%        newB = B;
%    end    
    
    
    if updateTransMat
        A = newA;
        B = newB;
        C = newC;
        D = newD;
    end
    
    if updateCovMat
        if updateTransMat
            R = 1/T*(y_y_0 - newC*Ey_x_0' - D*y_u_0');
            Q = 1/(T-1)*((Ex_x_0 - Exx_.start) - newA*Ex_x_1' - B*Ex_u_1');
        else
            cross = -Ey_x_0*C' + D*Eu_x_0*C' - y_u_0*D';
            R = 1/T*(y_y_0 + C*Ex_x_0*C' + D*u_u_0*D' + cross + cross');

            Eu_x_0x = Eu_x_0 - u_(:,T)*Ex_.end';
            cross = -Ex_u_1*B' + B*Eu_x_0x*A' - Ex_x_1*A';
            Q = 1/(T-1)*((Ex_x_0 - Exx_.start) + B*u_u_0_2*B' + A*(Ex_x_0 - Exx_.end)*A' + cross + cross');
        end
        
        %Q = (Q*(T-1) + T*0.00001*eye(hidsize))/(T-1);
    end
    
    %eig(R)
    %Q

    if updateInit
        pi_1 = Ex_.start;
        V_1 = Exx_.start - Ex_.start*Ex_.start';
    end

%% Stamp out some potential numerical problems

    V_1 = (V_1 + V_1')/2;
    R = (R + R')/2;
    Q = (Q + Q')/2;

%% Sanity Checks

    %If these get tripped you should increase k_lim or fix R and Q using some kind of heuristic that looks at their eigen-decompositions.
    
    %Alternatively, you could optimize for sqrt(R) and sqrt(Q) as is sometimes done to correct for this sort of problem

    if any(eig(R) < 0 )
         error('R developed negative eigenvalues')
    end

    %min(abs(eig(Q)))
    if any(eig(Q) < 0 )
        error('Q developed negative eigenvalues')
    end
    
%% Compute output measures
    if nargin >= 11 && mod(iter, sampleperiod) == 0
        if strcmp(method, 'exact')
            [expt, LL, err] = ExactEstep(y_, u_, pstruct( A, B, C, D, Q, R, pi_1, V_1 ) );
        elseif strcmp(method, 'SS')
            [expt, LL, err] = Steady_Exact_Estep(y_, u_, pstruct( A, B, C, D, Q, R, pi_1, V_1 ), 25 );
        else
            error( 'method string malformed');
        end
        
        LLhist(iter/sampleperiod) = LL;
        errhist(iter/sampleperiod) = sqrt(err/T);
    end    
    
    
end

params = pstruct( A, B, C, D, Q, R, pi_1, V_1 );

end

%% Computing the M-step objective (for debugging purposes)
function E = computeObj(y_y_0, u_u_0, u_u_0_2, y_u_0, u_, T, params, priors, expt)

    [A,B,C,D,Q,R,insize,outsize,hidsize,pi_1,V_1] = pextract( params );
    [Ex_x_0, Ex_x_1, Ey_x_0, Eu_x_0, Ex_u_1, Ex_, Exx_] = eextract( expt );
    if ~isempty( priors )
        [sigA, nQ, invVQ, nR, invVR] = rextract( priors );
    end
    
    invR = inv(R);
    invQ = inv(Q);
    invV_1 = inv(V_1);
    
    E = 0;
    
    cross = -Ey_x_0*C' + D*Eu_x_0*C' - y_u_0*D';
    E = E - vec(invR)'*vec(y_y_0 + C*Ex_x_0*C' + D*u_u_0*D' + 2*cross );
    
    Eu_x_0x = Eu_x_0 - u_(:,T)*Ex_.end';
    cross = -Ex_u_1*B' + B*Eu_x_0x*A' - Ex_x_1*A';
    E = E - vec(invQ)'*vec((Ex_x_0 - Exx_.start) + B*u_u_0_2*B' + A*(Ex_x_0 - Exx_.end)*A' + 2*cross);    
    
    E = E - vec(invV_1)'*vec(Exx_.start - 2*pi_1*Ex_.start' + pi_1*pi_1');
    E = E - T*logdet(R) - (T-1)*logdet(Q) - logdet(V_1);
    E = E - T*(outsize + hidsize)*log(2*pi);

    if ~isempty( priors )
        E = E - 1/(2*sigA^2)*vec(A - eye(hidsize))'*vec(A - eye(hidsize));
        E = E - vec(invR)'*vec(invVR) - (nR + outsize + 1)*logdet(R); %missing some constants
        E = E - vec(invQ)'*vec(invVQ) - (nQ + hidsize + 1)*logdet(Q); %missing some constants
        E = E - (2*hidsize^2*log(sigA) + hidsize^2*log(2*pi));
    end

    E = E/2;

    if imag(E) ~= 0
        error( 'Log of negative number while computing E (bad)' );
    end
end
