function [V_0_0, V_0_1, V_0_T, V_1_T, K, J, invS] = KalmanDoubling( A, Q, C, R, T )

%A is the state dynamics matrix
%Q is the dynamics noise covariance

%C is the prediction matrix
%R is the prediction noise covariance

numhid = size(A,1);

%% Forwards phase
Phi = A';
Psi = C'*inv(R)*C;
Theta = Q;

for t = 1:T
    mat = inv(eye(numhid) + Psi*Theta);
    incr = Phi'*Theta*mat*Phi;
    Theta = Theta + incr;
    if norm(vec(incr)) < 1e-8*norm(vec(Theta))
        break;
    end
    Psi = Psi + Phi*mat*Psi*Phi';
    Phi = Phi*mat*Phi;
end
%t

%% Asymptotic quantities from forwards phase

V_0_1 = Theta;
invS = inv(C*V_0_1*C' + R);
K = V_0_1*C'*invS;
V_0_0 = V_0_1 - K*C*V_0_1;

J = V_0_0*A'*inv(V_0_1);
D = V_0_0 - J*V_0_1*J';


%% Backwards phase
Phi = J';
Theta = D;

for t = 1:T
    incr = Phi'*Theta*Phi;
    Theta = Theta + incr;
    if norm(vec(incr)) < 1e-8*norm(vec(Theta))
        break;
    end
    Phi = Phi^2;
end
%t
V_0_T = Theta;


%% Cross phase
V_1_T = V_0_T*J'; %much simpler as in Bishop (but he's slightly wrong)
