function [expt, LL, perror] = Steady_Exact_Estep( y_, u_, params, edgesize )

T = size(y_,2);

[A,B,C,D,Q,R,insize,outsize,hidsize,pi_1,V_1] = pextract( params );

[V_0_0, V_0_1, V_0_T, V_1_T, K_s, J_s, invS_s] = KalmanDoubling( A, Q, C, R, 50 );


V0 = zeros( hidsize, hidsize, T );

x0 = zeros( hidsize, T );
x1 = zeros( hidsize, T );
xT = zeros( hidsize, T );

LL = 0;
perror = 0;

x1(:,1) = pi_1;
V1 = V_1;
for t = 1:edgesize
    if t > 1
        x1(:,t) = A*x0(:,t-1) + B*u_(:,t-1);
        V1 = A*V0(:,:,t-1)*A' + Q;
    end
    invS = inv(C*V1*C' + R);
    y_inov = y_(:,t) - C*x1(:,t) - D*u_(:,t);
    LL = LL - y_inov'*invS*y_inov / 2;
    LL = LL - outsize*log(2*pi)/2 + logdet(invS)/2;
    perror = perror + y_inov'*y_inov;
    
    K = V1*C'*invS;
    x0(:,t) = x1(:,t) + K*y_inov;
    V0(:,:,t) = V1 - K*C*V1;
end
for t = (edgesize+1):T
    if t > 1
        x1(:,t) = A*x0(:,t-1) + B*u_(:,t-1);
    end
    y_inov = y_(:,t) - C*x1(:,t) - D*u_(:,t);
    LL = LL - y_inov'*invS_s*y_inov / 2;
    perror = perror + y_inov'*y_inov;
    
    x0(:,t) = x1(:,t) + K_s*y_inov;
end
LL = LL - (outsize*log(2*pi)/2 - logdet(invS_s)/2)*(T-edgesize);


xT(:,T) = x0(:,T);
VT = V_0_0;
VcSum = zeros(hidsize, hidsize);
VTSum = VT;
for t = T:-1:(T-edgesize+1)
    %V1 = A*V0(:,:,t-1)*A' + Q;
    V1 = V_0_1;
    %J = V0(:,:,t-1)*A'*inv(V1);
    J = V_0_0*A'*inv(V1);
    VcSum = VcSum + VT*J';
    xT(:,t-1) = x0(:,t-1) + J*(xT(:,t) - x1(:,t));
    %VT = V0(:,:,t-1) + J*(VT - V1)*J';
    VT = V_0_0 + J*(VT - V1)*J';
    VTSum = VTSum + VT;
end

for t = (T-edgesize):-1:(edgesize+1)  %if you change this, change VTsum line as well
    xT(:,t-1) = x0(:,t-1) + J_s*(xT(:,t) - x1(:,t));
end
VTSum = VTSum + V_0_T*(T-2*edgesize);
VcSum = VcSum + V_0_T*J_s'*(T-2*edgesize);

VT = V_0_T;
for t = edgesize:-1:2
    V1 = A*V0(:,:,t-1)*A' + Q;
    J = V0(:,:,t-1)*A'*inv(V1);
    VcSum = VcSum + VT*J';
    xT(:,t-1) = x0(:,t-1) + J*(xT(:,t) - x1(:,t));
    VT = V0(:,:,t-1) + J*(VT - V1)*J';
    VTSum = VTSum + VT;
end




%% Compute final outputs of E-step

Exx_.start = xT(:,1)*xT(:,1)' + VT;
%Exx_.end = xT(:,T)*xT(:,T)' + V0(:,:,T);
Exx_.end = xT(:,T)*xT(:,T)' + V_0_0;

Ey_x_0 = y_*xT';

Eu_x_0 = u_*xT';

Ex_u_1 = xT(:,2:end)*u_(:,1:end-1)';

Ex_x_0 = xT*xT' + VTSum;

Ex_x_1 = xT(:,2:end)*xT(:,1:end-1)' + VcSum;


expt = estruct( Ex_x_0, Ex_x_1, Ey_x_0, Eu_x_0, Ex_u_1, xT, Exx_, x1 );
end

